/***********************************************************
 * dirlay.c -- a layout algorithm for directed graphs      *
 * Jun Yang, November 17, 1994				   *
 ***********************************************************/

/*--------------- References ---------------*/
/*
   This layout algorithm of directed graph is based 
   on recently developed techniques in this area.
   
   Reference:

   [1] Eades, P. and K. Sugiyama, "How to Draw a Directed Graph",
       Journal of Information PRocessing, Vol.13, No.4,
       pp.424-437, 1990.

   [2] Gansner, E.R., E. Koutsofios, S.C. North and K.-P. Vo,
       "A Technique for Drawing Directed Graphs", IEEE Trans.
       on Software Engineering, Vol.19, No.3, March 1993. pp.
       214-230.

   [3] Gansner, E.R., S.C. North and K.-P. Vo, "DAG-A Program
       that Draws Directed GRaphs", Software Practice and 
       Experience, 18(11) 1988, pp.1047-1062.

   [4] Rowe, L.A., M. Davis, E. Messinger, C. Meyer, C. Spirakis
       and A. Tuan, "A Browser for Directed Graphs", Software
       Practice and Experience, 17(1) 1987, pp.61-76.
*/

/*--------------- Header Files ---------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "swan.h"
#include "laydir.h"
#include "queue.h"
#include "queue.cpp"

#define MAXLAYER	50
#define EDGELENGTH	50.0
#define HOOKCONST	1.0 
#define WINFINITY	10000
#define EPSILONE	1.0
#define XYMARGIN	20.0
#define DEFNODEWIDTH	80
#define DEFLAYERDIST	50
#define MINXDIST	20
#define MAXITERATION	1
#define RRWEIGHT	8
#define RDWEIGHT	2
#define DDWEIGHT	1
#define LEFTBORDER	-8192
#define RIGHTBORDER	8192
#define DUMMYWIDTH	2
#define DUMMYHEIGHT	2

/*--------------- Types ---------------*/

struct LAYER {
    int		number ;
    double	avg_x ;
    _DI_NODELIST* node_list ;
} ; 

/*--------------- Globale Variables ---------------*/

_DI_Graph*	dg1 = NULL ; // a digraph 

Stack<_DI_Node*>* top_s = NULL ; // a queue storing the topological 
				 // order of the node in the graph
int		FAS_label ;
int		min_width = 0 ;
int		longest_path = 0 ;
struct LAYER	*layer_arr = NULL ;
_DI_Node**	layer_order[MAXLAYER] ;
_DI_Node**	layer_up_pri[MAXLAYER] ;
_DI_Node**	layer_down_pri[MAXLAYER] ;

//-------------------
//  DI_layout -- the main function for the layout algorithm
//___________________

bool DI_layout(BasicGraph* g_ptr)
{
	int	i ;

	if (g_ptr == NULL) {
	    _sw_errno = SWE_NOGRAPH ;
	    return FALSE ;
	}

	if (!g_ptr->isDirected()) 
	    return FALSE ;

	if (g_ptr->getNodeList()->getNodeCount() > NUMVERTEX) {
	    _sw_errno = SWE_MORENODES ;
	    return FALSE ;
	}

	if (g_ptr->getNodeList()->getNodeCount() == 0)
	    return TRUE ;

	dg1 = new _DI_Graph ;
	if (dg1 == NULL) {
	    _sw_errno = SWE_NOMEMORY ;
	    return FALSE ;
	}

	consAdjLists(g_ptr, dg1) ;	// construct adjacency matrix
	Simple_FAS(dg1) ;		// find a feedback arc set
	LongestPathLayer(dg1) ;		// layering the nodes
	addDummyNodes(dg1) ;		// break long edges by adding 
					// dummy nodes
	compMinPageWidth(dg1) ;
	nodeordering(dg1) ;		// order nodes on the same
					// layer
	assignDiPos(dg1) ;
	restoreEdgeDir(dg1) ;
	assignLabelRPos(g_ptr, dg1) ;
	_DI_calboxsize(g_ptr, dg1) ;

	// clean up
	delete top_s ;
	delete[] layer_arr ;
	delete[] layer_up_pri ;
	delete[] layer_down_pri ;
	for (i = 0 ; i < longest_path ; i++)
	    delete[] layer_order[i] ;

	return TRUE ; 
} /* End of DI_layout */

//-------------------
//  consAdjLists -- construct adjacency list of the graph  
//___________________

void consAdjLists(BasicGraph* g_ptr, _DI_Graph* di_ptr)
{
	bool		ok, ok2 ;
	BasicNode*	temp_node ;
	EDGELIST*	temp_el ;
	BasicEdge*	temp_edge ;
	NODELIST*	temp_nl ;
	NODENBLIST*	temp_nblist ;
	struct _nb_node* temp_nb ;
	_DI_Node*	di_node ;
	_DI_NODELIST*	di_list ;
	_DI_nbnode	*di_nb, *di_nb_in ;
	_DI_NBLIST*	di_nblist ;
	int		i, num_node ;
	BOXSIZE		temp_size ;
	struct NODEPAIR {
		_DI_NODEID	id ;
		_DI_Node*	addr ;
	} 		*pair_arr ; 
	
	temp_nl = g_ptr->getNodeList() ;
	ok = temp_nl->getFirstNode(temp_node) ;
	while (ok) {
	    di_ptr->insertNode(temp_node) ;
	    ok = temp_nl->getNextNode(temp_node) ;
	}

	num_node = temp_nl->getNodeCount() ;
	pair_arr = new NODEPAIR[num_node] ;
	i = 0 ;
	di_list = di_ptr->getNodeList() ;
	ok = temp_nl->getFirstNode(temp_node) ; 
	di_list->getFirstNode(di_node) ;
	while (ok) {
	    pair_arr[i].id = temp_node ;
	    pair_arr[i++].addr = di_node ;
	    ok = temp_nl->getNextNode(temp_node) ; 
	    di_list->getNextNode(di_node) ;
	}
	    
	ok = temp_nl->getFirstNode(temp_node) ;
	di_list->getFirstNode(di_node) ;
	while (ok) {
	    temp_nblist = temp_node->getNBList() ;
	    di_nblist = di_node->getNBList() ;
	    ok2 = temp_nblist->getFirstNode(temp_nb) ;
	    while (ok2) {
		di_nb = new _DI_nbnode ;
		for (i = 0 ; i < num_node ; i++) 
		    if (pair_arr[i].id == temp_nb->id) {
		   	di_nb->id = pair_arr[i].addr ;
			di_nb->edge_ptr = temp_nb->edge_ptr ;
			break ;
		    }
		di_nblist->insert(di_nb) ;
		di_nb_in = new _DI_nbnode ;
		di_nb_in->id = di_node ;
		di_nb_in->edge_ptr = temp_nb->edge_ptr ;
		di_nb->id->getNBList_in()->insert(di_nb_in) ;
		ok2 = temp_nblist->getNextNode(temp_nb) ;
	    }
	    temp_size = temp_node->getBoxSize() ;
	    di_node->setWidth(temp_size.w) ;
	    di_node->setHeight(temp_size.h) ;
	    ok = temp_nl->getNextNode(temp_node) ;
            di_list->getNextNode(di_node) ;
	}
	delete[] pair_arr ;

	// clear rpath of the edges 
	temp_el = g_ptr->getEdgeList() ;
	ok = temp_el->getFirstNode(temp_edge) ;
	while (ok) {
	    temp_edge->rpath->clear() ;
	    ok = temp_el->getNextNode(temp_edge) ;
	}
} /* End of consAdjLists */  

//-------------------
//  Simple_FAS -- find a Feedback Arc Set             
//___________________

void Simple_FAS(_DI_Graph* di_ptr)
{
	bool		ok ;
	_DI_Node	*temp_node ;
	_DI_NODELIST	*di_list ;

	FAS_label = 0 ;
	di_ptr->markVisitedFlags(FALSE) ;
	di_list = di_ptr->getNodeList() ;
	ok = di_list->getFirstNode(temp_node) ;
	while (ok) {
	    if (!temp_node->isVisited()) 
		simple_dfs(temp_node) ;
            ok = di_list->getNextNode(temp_node) ;
	}

	// reverse those up-arcs
	di_ptr->markVisitedFlags(FALSE) ;
	ok = di_list->getFirstNode(temp_node) ;
	while (ok) {
	    if (!temp_node->isVisited()) 
		double_dfs(temp_node) ;
            ok = di_list->getNextNode(temp_node) ;
	}
}

//-------------------
//  simple_dfs -- Depth-First Search of A Graph
//___________________

void simple_dfs(_DI_Node* di_node)
{
	_DI_NBLIST*	temp_nblist ;
	_DI_nbnode*	temp_nb ;
	bool		ok ;

	di_node->markVisited(TRUE) ;
	temp_nblist = di_node->getNBList() ;
	ok = temp_nblist->getFirstNode(temp_nb) ;
	while (ok) {
	    if (!temp_nb->id->isVisited())
	 	simple_dfs(temp_nb->id) ;
	    ok = temp_nblist->getNextNode(temp_nb) ;
	}
	FAS_label++ ;
	di_node->setLambda(FAS_label) ;
} /* End of simple_dfs */

//-------------------
//  double_dfs -- 
//___________________

void double_dfs(_DI_Node* di_node)
{
	_DI_NBLIST*	temp_nblist ;
	_DI_NBLIST*	temp_nblist_in ;
	_DI_nbnode*	temp_nb ;
	_DI_nbnode*	temp_nb_in ;
	_DI_Node*	next_node ;
	bool		ok, ok2 ;
	int		cur_lambda ;

	cur_lambda = di_node->getLambda() ;
	temp_nblist = di_node->getNBList() ;
	ok = temp_nblist->getFirstNode(temp_nb) ;
	while (ok) {
	    if (!temp_nb->id->isVisited()) {
		next_node = temp_nb->id ;
		if (cur_lambda < next_node->getLambda()) {
		    temp_nblist->remove(temp_nb) ;
		    temp_nb->id = di_node ;
		    temp_nb->fReversed = TRUE ;
		    next_node->getNBList()->insert(temp_nb) ;

		    temp_nblist_in = next_node->getNBList_in() ;
		    ok2 = temp_nblist_in->getFirstNode(temp_nb_in) ;
		    while (ok2) {
		  	if (temp_nb_in->id == di_node) { 
			    temp_nblist_in->remove(temp_nb_in) ;
			    temp_nb_in->id = next_node ;
			    temp_nb_in->fReversed = TRUE ;
			    di_node->getNBList_in()->insert(temp_nb_in) ;
			    break ;
			}
		        ok2 = temp_nblist_in->getNextNode(temp_nb_in) ;
		    }
		}
	 	double_dfs(next_node) ;
	    }
	    ok = temp_nblist->getNextNode(temp_nb) ;
	}
	di_node->markVisited(TRUE) ;
} /* End of double_dfs */

//-------------------
//  LongestPathLayer -- Layer the nodes accoding LongestPathLayer
//			algorithm
//___________________

void LongestPathLayer(_DI_Graph* di_ptr)
{
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;
	_DI_nbnode*	temp_nb ;
	_DI_NBLIST*	temp_nblist ;
	_DI_Node*	cur_node ;
	int		cur_layer ;
	bool		ok, ok2 ;
	
	toplogicalSort(di_ptr) ;

	temp_list = di_ptr->getNodeList() ;
	ok = temp_list->getFirstNode(temp_node) ;
	while (ok) {
	    temp_node->setLayer(0) ;
	    ok = temp_list->getNextNode(temp_node) ;
	}
	
	longest_path = 0 ;
	while (!top_s->isempty()) {
	    top_s->pop(cur_node) ;
	    cur_layer = cur_node->getLayer() + 1 ;

	    ok = temp_list->getFirstNode(temp_node) ;
	    while (ok) {
	        temp_nblist = temp_node->getNBList() ;
	        ok2 = temp_nblist->getFirstNode(temp_nb) ;
	        while (ok2) {
		    if (temp_nb->id == cur_node) {
		        if (temp_node->getLayer() < cur_layer) {
			    temp_node->setLayer(cur_layer) ;
			    if (cur_layer > longest_path)
				longest_path = cur_layer ;
			}		
			break ;
		    }
		    ok2 = temp_nblist->getNextNode(temp_nb) ;
	        }
		ok = temp_list->getNextNode(temp_node) ;
	    }
	}
	longest_path++ ;
} /* End of LongestPathLayer */

//-------------------
//  toplogicalSort -- Topological sort a directed graph          
//___________________

void toplogicalSort(_DI_Graph* di_ptr)
{
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;
	_DI_nbnode*	temp_nb ;
	_DI_NBLIST*	temp_nblist ;
	bool		ok ;

	int		 G_label = 0 ;
	Queue<_DI_Node*> *node_q = new Queue<_DI_Node*> ;

	top_s = new Stack<_DI_Node*> ;
	calIndegree(di_ptr) ;

	temp_list = di_ptr->getNodeList() ;
	ok = temp_list->getFirstNode(temp_node) ;
	while (ok) {
	    if (temp_node->Indegree == 0)
		node_q->add(temp_node) ;
	    ok = temp_list->getNextNode(temp_node) ;
	}
	
	while (!node_q->isempty()) {
	    node_q->remove(temp_node) ;
	    G_label++ ;
	    temp_node->top_label = G_label ;
	    top_s->push(temp_node) ;
	    temp_nblist = temp_node->getNBList() ;
	    ok = temp_nblist->getFirstNode(temp_nb) ;
	    while (ok) {
		temp_nb->id->Indegree-- ;
		if (temp_nb->id->Indegree == 0)
		    node_q->add(temp_nb->id) ;
		ok = temp_nblist->getNextNode(temp_nb) ;
	    }
	}
	delete node_q ;
} /* End of topologicalSort */

//-------------------
//  calIndegree -- calculate the in-degree of nodes              
//___________________

void calIndegree(_DI_Graph* di_ptr)
{
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;
	bool		ok ;

	temp_list = di_ptr->getNodeList() ;
	ok = temp_list->getFirstNode(temp_node) ;
	while (ok) {
	  temp_node->Indegree = temp_node->getNBList_in()->getNodeCount(); 
	  ok = temp_list->getNextNode(temp_node) ;
	}
} /* End of calIndegree */
	
//-------------------
//  calOutdegree -- calculate the out-degree of nodes              
//___________________

void calOutdegree(_DI_Graph* di_ptr)
{
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;
	bool		ok ;

	temp_list = di_ptr->getNodeList() ;
	ok = temp_list->getFirstNode(temp_node) ;
	while (ok) {
	    temp_node->Outdegree=temp_node->getNBList()->getNodeCount();
	    ok = temp_list->getNextNode(temp_node) ;
	}
} /* End of calOutdegree */

//-------------------
//  calInPriority -- calculate the in-nodes priority           
//___________________

void calInPriority(_DI_Graph* di_ptr)
{
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;
	_DI_nbnode*	temp_nb ; 
	_DI_NBLIST*	temp_nblist ;
	bool		ok, ok2 ;
	unsigned int	fType, fType2 ;

	temp_list = di_ptr->getNodeList() ;
	ok = temp_list->getFirstNode(temp_node) ;
	while (ok) {
	    temp_node->InPriority = 0 ;
	    ok = temp_list->getNextNode(temp_node) ;
	}

	ok = temp_list->getFirstNode(temp_node) ;
	while (ok) {
	    fType = temp_node->isDummy() ? 1 : 0 ;
	    temp_nblist = temp_node->getNBList_in() ;
	    ok2 = temp_nblist->getFirstNode(temp_nb) ;
	    while (ok2) {
		fType2 = temp_nb->id->isDummy() ? (fType|2) : fType ;
		switch (fType2) {
		case 0:
		    temp_node->InPriority += RRWEIGHT ;
		    break ;
		case 1:
		case 2:
		    temp_node->InPriority += RDWEIGHT ;
		    break ;
		case 3:
		    temp_node->InPriority += DDWEIGHT ;
		    break ;
		}
		ok2 = temp_nblist->getNextNode(temp_nb) ;
	    }
	    ok = temp_list->getNextNode(temp_node) ;
	}
} /* End of calInPriority */

//-------------------
//  calOutPriority -- calculate the out-nodes priority           
//___________________
	
void calOutPriority(_DI_Graph* di_ptr)
{
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;
	_DI_NBLIST*	temp_nblist ;
	_DI_nbnode*	temp_nb ; 
	bool		ok, ok2 ;
	unsigned int	fType, fType2 ;

	temp_list = di_ptr->getNodeList() ;
	ok = temp_list->getFirstNode(temp_node) ;
	while (ok) {
	    temp_nblist = temp_node->getNBList() ;
	    temp_node->OutPriority = 0 ;
	    if (temp_nblist != NULL) {
	        fType = temp_node->isDummy() ? 1 : 0 ;
		ok2 = temp_nblist->getFirstNode(temp_nb) ;
		while (ok2) {
		    fType2 = temp_nb->id->isDummy() ? (fType|2) : fType ;
		    switch (fType2) {
		    case 0:
		    	temp_node->OutPriority += RRWEIGHT ;
		    	break ;
		    case 1:
		    case 2:
		        temp_node->OutPriority += RDWEIGHT ;
		        break ;
		    case 3:
		        temp_node->OutPriority += DDWEIGHT ;
			break ;
		    }
		    ok2 = (*temp_nblist).getNextNode(temp_nb) ;
		}
	    }
	    ok = temp_list->getNextNode(temp_node) ;
	}
} /* End of calOutPriority */

//-------------------
//  sortLayerUpPriority --                                       
//___________________
	
void sortLayerUpPriority(void)
{
	register int	cur_layer, i, j, k ;
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;

	for (cur_layer = 0 ; cur_layer < longest_path ; cur_layer++) {
	    temp_list = layer_arr[cur_layer].node_list ;
	    k = temp_list->getNodeCount() ;
	    if (k == 0) 
		layer_up_pri[cur_layer] = NULL ;
	    else 
		layer_up_pri[cur_layer] = new _DI_Node*[k] ;
	    for (i = 0 ; i < k ; i++)
		layer_up_pri[cur_layer][i]=layer_order[cur_layer][i];

	    // sort down priorities
	    for (i = 0 ; i < k - 1 ; i++)
		for (j = i + 1 ; j < k ; j++)
		    if (layer_up_pri[cur_layer][i]->InPriority < 
			layer_up_pri[cur_layer][j]->InPriority) {
		   	temp_node = layer_up_pri[cur_layer][i] ;
			layer_up_pri[cur_layer][i] =
					layer_up_pri[cur_layer][j] ;
			layer_up_pri[cur_layer][j] = temp_node ;
		    }
	} 
} /* End of sortLayerUpPriority */ 

//-------------------
//  sortLayerDownPriority --                                       
//___________________
	
void sortLayerDownPriority(void)
{
	register int	cur_layer, i, j, k ;
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;

	for (cur_layer = 0 ; cur_layer < longest_path ; cur_layer++) {
	    temp_list = layer_arr[cur_layer].node_list ;
	    k = temp_list->getNodeCount() ;
	    if (k == 0) 
		layer_down_pri[cur_layer] = NULL ;
	    else 
		layer_down_pri[cur_layer] = new _DI_Node*[k] ;
	    for (i = 0 ; i < k ; i++)
		layer_down_pri[cur_layer][i]=layer_order[cur_layer][i];

	    // sort down priorities
	    for (i = 0 ; i < k - 1 ; i++)
		for (j = i + 1 ; j < k ; j++)
		    if (layer_down_pri[cur_layer][i]->OutPriority < 
			layer_down_pri[cur_layer][j]->OutPriority) {
		   	temp_node = layer_down_pri[cur_layer][i] ;
			layer_down_pri[cur_layer][i] =
					layer_down_pri[cur_layer][j] ;
			layer_down_pri[cur_layer][j] = temp_node ;
		    }
	} 
} /* End of sortLayerDownPriority */

//-------------------
//  addDummyNodes --                                               
//___________________
	
void addDummyNodes(_DI_Graph* di_ptr)
{
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;
	_DI_nbnode	*temp_nb, *temp_nb2, *temp_nb3 ;
	_DI_NBLIST*	temp_nblist ;
	int		cur_layer ;
	bool		ok, ok2 ;
	int		k ;
	_DI_Node*	temp_dummy ;
	_DI_Node*	pre_dummy ;
	_DI_nbnode	*dummy_nb, *dummy_nb_in ;
	int		layer_diff ;
	
	temp_list = di_ptr->getNodeList() ;
	ok = temp_list->getFirstNode(temp_node) ;
	while (ok) {
	    if (temp_node->isDummy())
		break ;
	    cur_layer = temp_node->getLayer() ;
	    temp_nblist = temp_node->getNBList() ;
	    ok2 = temp_nblist->getFirstNode(temp_nb) ;
	    while (ok2) {
		if (temp_nb->id->isDummy())
		    break ;
		layer_diff = cur_layer - temp_nb->id->getLayer() ;
	        if (layer_diff > 1) { // add layer_diff-1 dummy nodes
		    ok2 = temp_nblist->getNextNode(temp_nb2) ;
		    temp_nblist->remove(temp_nb) ;

		    // insert this dummy to node list
		    temp_dummy = new _DI_Node(NULL) ;
		    temp_dummy->setDummy(TRUE) ;
		    temp_dummy->setLayer(cur_layer-1) ;
		    temp_dummy->setWidth(DUMMYWIDTH) ;
		    temp_dummy->setHeight(DUMMYHEIGHT) ;
		    temp_list->insert(temp_dummy) ;
		    pre_dummy = temp_dummy ;

		    // insert this dummy to temp_node's nblist
		    dummy_nb = new _DI_nbnode ;
		    dummy_nb->id = temp_dummy ;
		    dummy_nb->fReversed = temp_nb->fReversed ;
		    dummy_nb->edge_ptr = temp_nb->edge_ptr ;
		    temp_nblist->insert(dummy_nb) ;

		    // insert temp_node to this dummy's nblist_in
		    dummy_nb_in = new _DI_nbnode ;
		    dummy_nb_in->id = temp_node ;
		    dummy_nb->fReversed = temp_nb->fReversed ;
		    dummy_nb_in->edge_ptr = temp_nb->edge_ptr ;
		    temp_dummy->getNBList_in()->insert(dummy_nb_in) ;

		    for (k = 2 ; k < layer_diff ; k++) {
		        temp_dummy = new _DI_Node(NULL) ;
		        temp_dummy->setDummy(TRUE) ;
		        temp_dummy->setLayer(cur_layer-k) ;
		        temp_dummy->setWidth(DUMMYWIDTH) ;
		        temp_dummy->setHeight(DUMMYHEIGHT) ;
		        temp_list->insert(temp_dummy) ;
			dummy_nb = new _DI_nbnode ;
			dummy_nb->id = temp_dummy ;
		        dummy_nb->fReversed = temp_nb->fReversed ;
			dummy_nb->edge_ptr = temp_nb->edge_ptr ;
			pre_dummy->getNBList()->insert(dummy_nb) ;

		        // insert temp_node to this dummy's nblist_in
		        dummy_nb_in = new _DI_nbnode ;
		        dummy_nb_in->id = pre_dummy ;
		        dummy_nb->fReversed = temp_nb->fReversed ;
		        dummy_nb_in->edge_ptr = temp_nb->edge_ptr ;
		        temp_dummy->getNBList_in()->insert(dummy_nb_in);

			pre_dummy = temp_dummy ;
		    }
		    dummy_nb = new _DI_nbnode ;
		    dummy_nb->id = temp_nb->id ;
		    dummy_nb->fReversed = temp_nb->fReversed ;
		    dummy_nb->edge_ptr = temp_nb->edge_ptr ;
		    pre_dummy->getNBList()->insert(dummy_nb) ;

	            // insert temp_node to this dummy's nblist_in
		    dummy_nb_in = new _DI_nbnode ;
		    dummy_nb_in->id = pre_dummy ;
		    dummy_nb->fReversed = temp_nb->fReversed ;
		    dummy_nb_in->edge_ptr = temp_nb->edge_ptr ;
		    temp_nb->id->getNBList_in()->insert(dummy_nb_in);

		    if ((temp_nb3 = temp_nb->id->searchNB_in(temp_node))
							!= NULL)
			temp_nb->id->getNBList_in()->remove(temp_nb3) ;
		    temp_nb = temp_nb2 ;
		}
		else 
		    ok2 = temp_nblist->getNextNode(temp_nb) ;
	    }	
	    ok = temp_list->getNextNode(temp_node) ;
	}
} /* End of addDummyNodes */

//-------------------
//  compMinPageWidth --                                            
//___________________
	
void compMinPageWidth(_DI_Graph* di_ptr)
{
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;
	bool		ok ;
	int		i, max_node ;
	int		cur_layer ;

	layer_arr = new LAYER[longest_path] ;
	for (i = 0 ; i < longest_path ; i++) {
	    layer_arr[i].number = 0 ;
	    layer_arr[i].avg_x = 0.0 ;
	    layer_arr[i].node_list = new _DI_NODELIST ;
	}

	temp_list = (*di_ptr).getNodeList() ;
	ok = (*temp_list).getFirstNode(temp_node) ;
	while (ok) {
	    cur_layer = (*temp_node).getLayer() ;
	    (layer_arr[cur_layer].number)++ ;
	    (*(layer_arr[cur_layer].node_list)).insert(temp_node) ;
	    ok = (*temp_list).getNextNode(temp_node) ;
	}

	max_node = layer_arr[0].number ;
	for (i = 1 ; i < longest_path ; i++)
	    if (layer_arr[i].number > max_node)
		max_node = layer_arr[i].number ;

	min_width = DEFNODEWIDTH * (max_node+1) ;	

}

void nodeordering(_DI_Graph* di_ptr)
{
	init_order(di_ptr) ;
	barycenterSort(di_ptr) ;
	transpose(di_ptr) ;
} // end of nodeordering

void barycenterSort(_DI_Graph* di_ptr)
{
	int		i, j;
	_DI_Node*	temp_node ;
	int		cur_layer, cur_num, nb_num ;
	_DI_NODELIST*	temp_list ;	
	_DI_NBLIST*	temp_nblist ;
	_DI_nbnode*	temp_nb ;
	int		sum_x ;
	bool		ok, ok2 ;

	init_pos() ;

	for (cur_layer = 1; cur_layer < longest_path; cur_layer++) {
	    cur_num = 
		(*(layer_arr[cur_layer].node_list)).getNodeCount();
	    temp_list = layer_arr[cur_layer].node_list ;
	    ok = (*temp_list).getFirstNode(temp_node) ;
	    while (ok) {
	 	temp_nblist = (*temp_node).getNBList() ;
		sum_x = 0 ;
		ok2 = (*temp_nblist).getFirstNode(temp_nb) ;
		while (ok2) {
		    if ((*(temp_nb->id)).getLayer() == cur_layer - 1) 
		        sum_x += (*(temp_nb->id)).getX() ;
		    ok2 = (*temp_nblist).getNextNode(temp_nb) ;
		}
		nb_num = (*temp_nblist).getNodeCount() ;
		if (nb_num > 0) 
	            temp_node->avg_x = (double)sum_x / nb_num ;
		else 
		    temp_node->avg_x = -1 ;
		ok = (*temp_list).getNextNode(temp_node) ;
	    }

	    // sort according to Bary center
	    for (i = 0 ; i < cur_num - 1 ; i++) 
		if (layer_order[cur_layer][i]->avg_x != -1) {
		  for (j = i + 1 ; j < cur_num ; j++)  
		    if (layer_order[cur_layer][j]->avg_x != -1) 
		      if (layer_order[cur_layer][i]->avg_x > 
			  layer_order[cur_layer][j]->avg_x) {
			  temp_node = layer_order[cur_layer][i] ;
			  temp_node->order = j ;
			  layer_order[cur_layer][j]->order = i ;
			  layer_order[cur_layer][i] = 
					layer_order[cur_layer][j] ;
			  layer_order[cur_layer][j] = temp_node ;
		      }
		}
	    
	    // reset cur_layer's node's position according to new order
	    set_one_layer_pos(cur_layer) ;
	}
}

void init_order(_DI_Graph* di_ptr)
{
	register int	k, cur_layer ;
	int		prev_layer, prev_num ;
	int		cur_k, cur_num ;
	_DI_Node	*temp_node, *cur_parent ;
	_DI_NODELIST	*temp_list ;
	_DI_NBLIST	*temp_nblist ;
	bool		ok ;
	_DI_nbnode	*temp_nb ;	

	(*di_ptr).markVisitedFlags(FALSE) ;

	cur_layer = longest_path - 1 ;
	temp_list = layer_arr[cur_layer].node_list ;
	cur_num = (*temp_list).getNodeCount() ;
	layer_order[cur_layer] = new _DI_Node*[cur_num] ;
	k = 0 ;  
	ok = (*temp_list).getFirstNode(temp_node) ;
	while (ok) {
	    if (!(*temp_node).isVisited()) {
	        layer_order[cur_layer][k] = temp_node ;
	        temp_node->order = k ;
	        k++ ;
		(*temp_node).markVisited(TRUE) ;
	    }
	    ok = (*temp_list).getNextNode(temp_node) ;
	}

	prev_layer = cur_layer ;
	prev_num = cur_num ;
	cur_layer-- ;
	while (cur_layer >= 0) {
	  cur_num = (*(layer_arr[cur_layer].node_list)).getNodeCount() ;
	  layer_order[cur_layer] = new _DI_Node*[cur_num] ;
	  cur_k = 0 ;
	  for (k = 0 ; k < prev_num ; k++) {
	    cur_parent = layer_order[prev_layer][k] ;
	    temp_nblist = (*cur_parent).getNBList() ;
	    ok = (*temp_nblist).getFirstNode(temp_nb) ;
	    while (ok) {
	      if (!(*(temp_nb->id)).isVisited()) {
	        if ((*(temp_nb->id)).getLayer() == cur_layer) {
                  layer_order[cur_layer][cur_k] = temp_nb->id ;	
		  temp_nb->id->order = cur_k ;
		  cur_k++ ;
		  (*(temp_nb->id)).markVisited(TRUE) ;
	        }
	      }
	      ok = (*temp_nblist).getNextNode(temp_nb) ;
	    }
	  }
	  if (cur_k < cur_num) { 
	    temp_list = layer_arr[cur_layer].node_list ;
	    ok = (*temp_list).getFirstNode(temp_node) ;
	    while (ok) {
		if (!(*temp_node).isVisited()) {
		    layer_order[cur_layer][cur_k] = temp_node ;
		    temp_node->order = cur_k++ ;
		    (*temp_node).markVisited(TRUE) ;
		}
		ok = (*temp_list).getNextNode(temp_node) ;
	    }
	  }
	  if (cur_k != cur_num) 
	      exit(-1) ;
		
	  prev_layer = cur_layer ;
	  prev_num = cur_num ;
	  cur_layer-- ;
	}
} /* end of init_order */

void transpose(_DI_Graph* di_ptr)
{
	register int	i, k ;
	bool improved = TRUE ;
	_DI_Node *temp_node ;
	int	cur_num ;

	while (improved) {
	    improved = FALSE ;
	    for (i = 1 ; i < longest_path ; i++) {
	      cur_num = (*(layer_arr[i].node_list)).getNodeCount() ;
	      for (k = 0 ; k < cur_num - 2 ; k++) {
	        if (crossing_down(layer_order[i][k], layer_order[i][k+1])
		    > crossing_down(layer_order[i][k+1],
			       layer_order[i][k])) {
		    improved = TRUE ;
		    temp_node = layer_order[i][k] ;
		    temp_node->order = k+1 ;
		    layer_order[i][k+1]->order = k ;
		    layer_order[i][k] = layer_order[i][k+1] ;
		    layer_order[i][k+1] = temp_node ;
		}
	      }
	    }
	}
} /* end of transpose */
	 	    	
int crossing_down(_DI_Node* v, _DI_Node* w)
{
	register int	k, cur_order ;
	int		counter = 0 ;
	int		*order_arr, nb_num1, cur_layer ;
	_DI_NBLIST	*nblist ;
	_DI_nbnode	*temp_nb ;
	bool		ok ;

	if ((*v).getLayer() != (*w).getLayer())
	    return -1 ;
	cur_layer = (*v).getLayer() - 1 ;

	nblist = (*v).getNBList() ;
	nb_num1 = (*nblist).getNodeCount() ;
	if (nb_num1 == 0)
	    return 0 ;

	order_arr = new int[nb_num1] ;
	k = 0 ;
	ok = (*nblist).getFirstNode(temp_nb) ;
	while (ok) {
	  if ((*temp_nb->id).getLayer() == cur_layer) 
	    order_arr[k++] = temp_nb->id->order ;
	    ok = (*nblist).getNextNode(temp_nb) ;
	}
	nb_num1 = k ;
	qsort(order_arr, nb_num1, sizeof(int), cmp_int) ;

	nblist = (*w).getNBList() ;
	if ((*nblist).getNodeCount() == 0)
	    return 0 ;
	ok = (*nblist).getFirstNode(temp_nb) ;
	while (ok) {
	    cur_order = temp_nb->id->order ;
	    for (k = 0 ; k < nb_num1 ; k++) 
		if (order_arr[k] > cur_order) {
		    counter += nb_num1 - k ;
		    break ;
		}
	    ok = (*nblist).getNextNode(temp_nb) ;
	}
	delete[] order_arr ;
	return counter ;
} /* end of crossing_down */

// assign node's position according to its order in the layer
void init_pos(void)
{
	register int	cur_layer ;
	for (cur_layer = 0 ; cur_layer < longest_path ; cur_layer++) 
	    set_one_layer_pos(cur_layer) ;
} /* end of init_pos */

void set_one_layer_pos(int layer)
{
	register int 	cur_x, k, cur_num ;
	_DI_Node 	*cur_node ;

	cur_x = 0 ;
	cur_num = 
            (*(layer_arr[layer].node_list)).getNodeCount() ;
	for (k = 0 ; k < cur_num ; k++) {
	    cur_node = layer_order[layer][k] ;
	    (*cur_node).setX(cur_x) ;
	    (*cur_node).setY((longest_path-layer)*DEFLAYERDIST) ;
	    cur_x += (*cur_node).getWidth() + MINXDIST ;
	}
} /* end of set_one_layer_pos */

void assignDiPos(_DI_Graph* di_ptr)
{
	register int 	i, j, k ;
	int		ii ;
	int		cur_order, left_order, right_order ;
	int		cur_x, sum_nb_x, median_x, min_x ;
	int		temp_x, offset ;
	int		left_limit, right_limit ;
	int		cur_l, cur_n, num_n, num_nb ;
	LOCATION	temp_pos ;
	_DI_Node	*cur_node ;
	_DI_NODELIST*	temp_list ;
	bool		ok ;
	_DI_NBLIST	*temp_nblist ;
	_DI_nbnode	*temp_nb ;

	// cal indegree and out degree
	calInPriority(di_ptr) ;
	calOutPriority(di_ptr) ;
	sortLayerUpPriority() ;
	sortLayerDownPriority() ;
	init_pos() ;
/*
	for (i = 0 ; i < longest_path ; i++) {
	    cur_x = 0 ;
	    for (j = 0 ; j < (*(layer_arr[i].node_list)).getNodeCount();
		 j++) {
		cur_node = layer_order[i][j] ;
		(*cur_node).setX(cur_x) ;
		(*cur_node).setY((longest_path-i)*DEFLAYERDIST) ;
		cur_x += (*cur_node).getWidth() + MINXDIST ;
	    }
	} 
*/

	
	// iteratively modify the position
	for (ii = 0 ; ii < MAXITERATION ; ii++) {

	// downward, actually bottom up
	for (cur_l = 1 ; cur_l < longest_path ; cur_l++) {
	  num_n = (*(layer_arr[cur_l].node_list)).getNodeCount() ;
 	  for (cur_n = 0 ; cur_n < num_n ; cur_n++) {
	    cur_node = layer_down_pri[cur_l][cur_n] ;

	    // find cur_node's order in the layer
/*
	    for (k = 0 ; k < num_n ; k++)
		if (layer_order[cur_l][k] == cur_node) {
		    cur_order = k ;
		    break ;
		}
*/
	    cur_order = cur_node->order ;

	    // find cur_node's median
	    sum_nb_x = 0 ;
	    temp_nblist = (*cur_node).getNBList() ;
	    ok = (*temp_nblist).getFirstNode(temp_nb) ;
	    while (ok) {
		if ((*(temp_nb->id)).getLayer() == cur_l - 1) 
		    sum_nb_x += (*(temp_nb->id)).getX() ;
		ok = (*temp_nblist).getNextNode(temp_nb) ;
	    }
	    num_nb = (*temp_nblist).getNodeCount() ;
	    if (num_nb > 0) 
	        median_x = sum_nb_x / num_nb ;
	    else
		continue ; // process next node

	    // find cur_node's moving range
	    left_order = cur_order - 1 ;
	    right_order = cur_order + 1 ;
	    while (left_order >= 0 && 
	      layer_order[cur_l][left_order]->OutPriority <
				    cur_node->OutPriority)
		left_order-- ;
	    while (right_order < num_n && 
	      layer_order[cur_l][right_order]->OutPriority <
				     cur_node->OutPriority)
		right_order++ ;
	    if (left_order < 0)
		left_limit = LEFTBORDER ;
	    else 
		left_limit = (*layer_order[cur_l][left_order]).getX() + 
			     (*layer_order[cur_l][left_order]).getWidth() +
			     MINXDIST ;
	    if (right_order >= num_n)
		right_limit = RIGHTBORDER ;
	    else 
		right_limit = (*layer_order[cur_l][right_order]).getX() -
				(*cur_node).getWidth() - MINXDIST ;

	    offset = right_limit - left_limit ;
	    if (offset < 0) {
		for (k = cur_order+1 ; k < num_n ; k++) {
		    temp_x = (*layer_order[cur_l][k]).getX() ;
		    (*layer_order[cur_l][k]).setX(temp_x - offset) ;
		}
	        right_limit = left_limit ;
	    }

	    // assign X pos to cur_node
	    if (median_x < left_limit) 
		(*cur_node).setX(left_limit) ;
	    else if (median_x > right_limit)
		(*cur_node).setX(right_limit) ;
	    else
		(*cur_node).setX(median_x) ;
	  } // end for cur_n	
	} // end of for cur_l

	// upward, actually top down 
	for (cur_l = longest_path-1 ; cur_l >= 0 ; cur_l--) {
	  num_n = (*(layer_arr[cur_l].node_list)).getNodeCount() ;
 	  for (cur_n = 0 ; cur_n < num_n ; cur_n++) {
	    cur_node = layer_up_pri[cur_l][cur_n] ;

	    // find cur_node's order in the layer
	    for (k = 0 ; k < num_n ; k++)
		if (layer_order[cur_l][k] == cur_node) {
		    cur_order = k ;
		    break ;
		}

	    // find cur_node's median
	    sum_nb_x = 0 ;
	    temp_nblist = (*cur_node).getNBList_in() ;
	    ok = (*temp_nblist).getFirstNode(temp_nb) ;
	    while (ok) {
		if ((*(temp_nb->id)).getLayer() == cur_l + 1) 
		    sum_nb_x += (*(temp_nb->id)).getX() ;
		ok = (*temp_nblist).getNextNode(temp_nb) ;
	    }
	    num_nb = (*temp_nblist).getNodeCount() ;
	    if (num_nb > 0) 
	        median_x = sum_nb_x / num_nb ;
	    else
		continue ; // process next node

	    // find cur_node's moving range
	    left_order = cur_order - 1 ;
	    right_order = cur_order + 1 ;
	    while (left_order >= 0 && 
	      layer_order[cur_l][left_order]->InPriority<cur_node->InPriority)
		left_order-- ;
	    while (right_order < num_n && 
	      layer_order[cur_l][right_order]->InPriority<cur_node->InPriority)
		right_order++ ;
	    if (left_order < 0)
		left_limit = LEFTBORDER ;
	    else 
		left_limit = (*layer_order[cur_l][left_order]).getX() + 
			     (*layer_order[cur_l][left_order]).getWidth() +
			     MINXDIST ;
	    if (right_order >= num_n)
		right_limit = RIGHTBORDER ;
	    else 
		right_limit = (*layer_order[cur_l][right_order]).getX() -
				(*cur_node).getWidth() - MINXDIST ;

	    offset = right_limit - left_limit ;
	    if (offset < 0) {
		for (k = cur_order+1 ; k < num_n ; k++) {
		    temp_x = (*layer_order[cur_l][k]).getX() ;
		    (*layer_order[cur_l][k]).setX(temp_x - offset) ;
		}
	        right_limit = left_limit ;
	    }

	    // assign X pos to cur_node
	    if (median_x < left_limit) 
		(*cur_node).setX(left_limit) ;
	    else if (median_x > right_limit)
		(*cur_node).setX(right_limit) ;
	    else
		(*cur_node).setX(median_x) ;
	  } // end for cur_n	
	} // end of for cur_l
	} // end of ii

	// find min X 
	min_x = 0 ;
	temp_list = (*di_ptr).getNodeList() ;
	ok = (*temp_list).getFirstNode(cur_node) ;
	while (ok) {
	    if ((*cur_node).getX() < min_x)
		min_x = (*cur_node).getX() ;
	    ok = (*temp_list).getNextNode(cur_node) ;
	}
	if (min_x < 0) {
	    ok = (*temp_list).getFirstNode(cur_node) ;
	    while (ok) {
	       (*cur_node).setX((*cur_node).getX()-min_x) ;
	       ok = (*temp_list).getNextNode(cur_node) ;
	    }
	}

	// assign node positions back
        temp_list = (*di_ptr).getNodeList() ;
        ok = (*temp_list).getFirstNode(cur_node) ;
        while (ok) {
            if (!(*cur_node).isDummy()) {
                temp_pos.x = (*cur_node).getX() ;
                temp_pos.y = (*cur_node).getY() ;
                (*(*cur_node).getID()).setRPos(temp_pos) ;
            }
            ok = (*temp_list).getNextNode(cur_node) ;
        }

	// set edge positions
	(*di_ptr).markVisitedFlags(FALSE) ;

	temp_list = (*di_ptr).getNodeList() ;
	ok = (*temp_list).getFirstNode(cur_node) ;
	while (ok) {
	    if (!(*cur_node).isVisited() &&
		!(*cur_node).isDummy()) 
		edgepos_dfs(cur_node) ;
	    ok = (*temp_list).getNextNode(cur_node) ;
	}

} /* end of assignDiPos */
 
void edgepos_dfs(_DI_Node* di_node)
{
	_DI_NBLIST*	temp_nblist ;
	_DI_nbnode*	temp_nb ;
	bool		ok ;
	LOCATION	temp_pos, temp_pos2 ;

	(*di_node).markVisited(TRUE) ;
	temp_pos.x = (*di_node).getX() ;
	temp_pos.y = (longest_path - (*di_node).getLayer())
			* DEFLAYERDIST ;
	temp_nblist = (*di_node).getNBList() ;
	ok = (*temp_nblist).getFirstNode(temp_nb) ;
	while (ok) {
		if (!(*di_node).isDummy())
		    (*(temp_nb->edge_ptr->rpath)).insert(temp_pos) ;
		temp_pos2.x = (*(temp_nb->id)).getX() ;
		temp_pos2.y = (longest_path - (*(temp_nb->id)).getLayer())
			* DEFLAYERDIST ;
		(*(temp_nb->edge_ptr->rpath)).insert(temp_pos2) ;   
		if (!(*(temp_nb->id)).isVisited()) 
	 	    edgepos_dfs(temp_nb->id) ;
	    ok = (*temp_nblist).getNextNode(temp_nb) ;
	}
}

void restoreEdgeDir(_DI_Graph* di_ptr)
{
	LOCATION	temp_pos ;
	_DI_Node*	temp_node ;
	_DI_NODELIST*	temp_list ;
	bool		ok, ok2, ok3 ;
	_DI_NBLIST*	temp_nblist ;
	_DI_nbnode*	temp_nb ;
	CURVE		*new_path = new CURVE ;
	CURVE		*old_path = NULL ;
	Stack<LOCATION> *temp_stack = new Stack<LOCATION> ;

	temp_list = (*di_ptr).getNodeList() ;
	ok = (*temp_list).getFirstNode(temp_node) ;
	while (ok) {
	    if (!(*temp_node).isDummy()) {
		temp_nblist = (*temp_node).getNBList() ;
		ok2 = (*temp_nblist).getFirstNode(temp_nb) ;
		while (ok2) {
		    if (temp_nb->fReversed) { 
			// reverse the edge direction back 
			old_path = temp_nb->edge_ptr->rpath ;
			ok3 = (*old_path).getFirstNode(temp_pos) ;
			while (ok3) {
			    (*temp_stack).push(temp_pos) ;
			    ok3 = (*old_path).getNextNode(temp_pos) ;
			}
			while (!(*temp_stack).isempty()) {
			    (*temp_stack).pop(temp_pos) ;
			    (*new_path).insert(temp_pos) ;
			}
			(*(temp_nb->edge_ptr)).setRPath(new_path) ;
			(*new_path).clear() ;
		    }
		    ok2 = (*temp_nblist).getNextNode(temp_nb) ;
		}
	    }
	    ok = (*temp_list).getNextNode(temp_node) ;
	}
	delete temp_stack ;
}

bool assignLabelRPos(BasicGraph* g_ptr, _DI_Graph* di_ptr)
{
	BasicNode	*s_node, *e_node ;
	_DI_Node	*s_di, *e_di ;
	BasicEdge	*temp_edge ;
	EDGELIST	*edge_list ;
        LOCATION        spos, epos ;
        EDGEINFO        *temp_info ;
	bool		ok ;

	if (g_ptr == NULL) 
	    return FALSE ;
	if (di_ptr == NULL) 
	    return FALSE ;
	edge_list = (*g_ptr).getEdgeList() ;
	ok = (*edge_list).getFirstNode(temp_edge) ;
	while (ok) {
	    s_node = (*temp_edge).getSNode() ;
	    e_node = (*temp_edge).getENode() ;
	    if ((s_di = (*di_ptr).searchNode(s_node)) == NULL)
		return FALSE ;
	    if ((e_di = (*di_ptr).searchNode(e_node)) == NULL)
		return FALSE ;
	    spos = (*s_node).getRPos() ;
	    epos = (*e_node).getRPos() ;
	    temp_info = (*temp_edge).getAttr() ;
	
	    if ((*s_di).getLayer() > (*e_di).getLayer()) {
		if ((*(*s_di).getNBList()).getNodeCount() < 
		    (*(*e_di).getNBList_in()).getNodeCount()) 
		    temp_info->label_ppos = spos ;
		else 
		    temp_info->label_ppos = epos ;
	    }
	    else {
		if ((*(*e_di).getNBList()).getNodeCount() < 
		    (*(*s_di).getNBList_in()).getNodeCount()) 
		    temp_info->label_ppos = epos ;
		else
		    temp_info->label_ppos = spos ;
	    }
	    ok = (*edge_list).getNextNode(temp_edge) ;
	}
	return TRUE ;
} // end of assignLabelRPos
	
bool _DIGRAPH_edge_ppath(BasicGraph* g_ptr)
{
        bool         ok ;
        BasicEdge*      temp_edge ;
        BasicNode       *s_node, *e_node ;
        EDGELIST*       temp_edgelist ;
        CURVE           *ppath, *rpath ;
        LOCATION        spos, epos, pos1, pos2, pos3, lc_pos ;
	LOCATION	pre_pos, cur_pos ; 
        BOXSIZE         ssize, esize ;
	BasicLayoutComp *lc1 ;

        temp_edgelist = (*g_ptr).getEdgeList() ;
        ok = (*temp_edgelist).getFirstNode(temp_edge) ;
        while (ok) {
            s_node = (*temp_edge).getSNode() ;
            e_node = (*temp_edge).getENode() ;
            lc1 = (*s_node).getLayoutComp() ;
            ssize = ((*s_node).getAttr())->size ;
            esize = ((*e_node).getAttr())->size ;
            ppath = (*temp_edge).getPPath() ;
	    (*ppath).clear() ;
	    rpath = (*temp_edge).getRPath() ;
	    spos = (*s_node).getPPos() ;
	    epos = (*e_node).getPPos() ;
	    if (spos != UnsetPos && epos != UnsetPos) {
            if ((*rpath).getNodeCount() == 2) {
                                                // two nodes edge
                if ((*temp_edge)._intersections((*s_node).getPPos(),
                    (*s_node).getType(), (*s_node).getBoxSize(),
                    (*e_node).getPPos(), (*e_node).getType(),
                    (*e_node).getBoxSize(), pos1, pos2)) {
                    (*ppath).insert(pos1) ;
                    (*ppath).insert(pos2) ;
                }
	    }
            else if ((*rpath).getNodeCount() > 2) 
                // there are dummy nodes in between

                switch ((*lc1).getNodeType()) {
                case BOX:
            	    spos = (*s_node).getPPos() ;
            	    epos = (*e_node).getPPos() ;
                    lc_pos = (*lc1).getPPos() ;
                    (*rpath).getFirstNode(pos1) ; // skip first pos
                    (*rpath).getNextNode(pos2) ;
                    pos2.x += lc_pos.x ;
                    pos2.y += lc_pos.y ;
                    line_box_crosspoint(pos2, spos, ssize.w, ssize.h, pos1) ;
                    (*ppath).insert(pos1) ;
		    pre_pos = pos1 ;
		    cur_pos = pos2 ;
                    ok = (*rpath).getNextNode(pos2) ;
                    while (ok) {
                        (*ppath).insert(cur_pos) ;
                        pre_pos = cur_pos ;
                        pos2.x += lc_pos.x ;
                        pos2.y += lc_pos.y ;
			cur_pos = pos2 ;
                        ok = (*rpath).getNextNode(pos2) ;
                    }
                    line_box_crosspoint(pre_pos, epos, esize.w, esize.h, pos2) ;
                    (*ppath).insert(pos2) ;
                    break ;

                case CIRCLE:
            	    spos = (*s_node).getPPos() ;
            	    epos = (*e_node).getPPos() ;
                    lc_pos = (*lc1).getPPos() ;
                    (*rpath).getFirstNode(pos1) ; // skip first pos
                    (*rpath).getNextNode(pos1) ;
                    pos1.x += lc_pos.x ;
                    pos1.y += lc_pos.y ;
                    line_ellipse_crosspoint(pos1, spos, ssize.w,
						ssize.h, pos3) ;
                    (*ppath).insert(pos3) ;
		    pre_pos = pos3 ;
		    cur_pos = pos1 ;
                    ok = (*rpath).getNextNode(pos2) ;
                    while (ok) {
                        (*ppath).insert(cur_pos) ;
                        pre_pos = cur_pos ;
                        pos2.x += lc_pos.x ;
                        pos2.y += lc_pos.y ;
			cur_pos = pos2 ;
                        ok = (*rpath).getNextNode(pos2) ;
                    }
                    line_ellipse_crosspoint(pre_pos, epos, esize.w,
						esize.h, pos2) ;
                    (*ppath).insert(pos2) ;
                    break ;
                default:
                    break ;
                }
	    }
            ok = (*temp_edgelist).getNextNode(temp_edge) ;
	}
	assignLabelPPos(g_ptr) ;
	return ok;  //This function must return a value I hope this is the right one
	            //1/26/97 jrf
}

bool assignLabelPPos(BasicGraph* g_ptr)
{
	BasicNode	*s_node, *e_node ;
	BasicEdge	*temp_edge ;
	EDGELIST	*edge_list ;
        LOCATION        pos1, pos2, srpos, erpos ;
        EDGEINFO        *temp_info ;
        BOXSIZE         temp_size ;
        LABEL           temp_label ;
	int		len ;
	bool		ok ;

	if (g_ptr == NULL) 
	    return FALSE ;
	edge_list = (*g_ptr).getEdgeList() ;
	ok = (*edge_list).getFirstNode(temp_edge) ;
	while (ok) {
	    s_node = (*temp_edge).getSNode() ;
	    e_node = (*temp_edge).getENode() ;
	    srpos = (*s_node).getRPos() ;
	    erpos = (*e_node).getRPos() ;

            temp_label = (*temp_edge).getLabel() ;
	    if (temp_label == NULL)
		len = 0 ;
	    else 
		len = strlen(temp_label) ;
            temp_size.w = CURFONTWIDTH * len ;
            temp_size.h = CURFONTHEIGHT ;
            temp_info = (*temp_edge).getAttr() ;

	    if (temp_info->label_ppos == srpos) {
		if (!(*temp_edge).getFirst2PPath(pos1, pos2))
		    return FALSE ;
	    }
	    else if (temp_info->label_ppos == erpos) {
		if (!(*temp_edge).getLast2PPath(pos1, pos2))
		    return FALSE ;
	    }
	    else
		return FALSE ;
            _sw_segedgepos(pos1, pos2, temp_size, temp_info) ;
	    ok = (*edge_list).getNextNode(temp_edge) ;
	} // end of while ok
	return TRUE ;
} // end of assignLabelPos

bool _DI_calboxsize(BasicGraph* g_ptr, _DI_Graph* di_ptr)
{
        BasicLayoutComp *temp_lc, *root_lc ;
        LCNBLIST        *temp_lclist ;
        BOXSIZE         temp_size;
        _DI_NODELIST        *temp_nodelist ;
        _DI_Node        *temp_node ;
        bool         ok ;

        root_lc = (*g_ptr).getRootLC() ;
        temp_lclist = (*root_lc).getChLCList() ;
        if (!(*temp_lclist).getFirstNode(temp_lc))
            return FALSE ;

	if (di_ptr == NULL)
	    return TRUE ;

        temp_size.w = temp_size.h = 0 ;
        temp_nodelist = (*di_ptr).getNodeList() ;
        ok = (*temp_nodelist).getFirstNode(temp_node) ;
        while (ok) {
	    if ((*temp_node).getX() + (*temp_node).getWidth() >
		temp_size.w)
		temp_size.w = (*temp_node).getX() + 
			      (*temp_node).getWidth() ;
	    if ((*temp_node).getY() + (*temp_node).getHeight() >
		temp_size.h)
		temp_size.h = (*temp_node).getY() + 
			      (*temp_node).getHeight() ;
            ok = (*temp_nodelist).getNextNode(temp_node) ;
        }
        (*temp_lc).setBoxSize(temp_size) ;
        (*root_lc).setBoxSize(temp_size) ;
        return TRUE ;
}

/***********************************************************
 * End of dirlay.c                          		   * 
 ***********************************************************/
